/// <summary>Common definitions shared between attribute values and response headers</summary>
[<AutoOpen>]
module Giraffe.Htmx.Common

/// <summary>Serialize a list of key/value pairs to JSON (very rudimentary)</summary>
/// <param name="pairs">The key/value pairs to be serialized to JSON</param>
/// <returns>A string with the key/value pairs serialized to JSON</returns>
let internal toJson (pairs: (string * string) list) =
    pairs
    |> List.map (fun pair -> sprintf "\"%s\": \"%s\"" (fst pair) ((snd pair).Replace ("\"", "\\\"")))
    |> String.concat ", "
    |> sprintf "{ %s }"

/// <summary>Convert a boolean to lowercase <tt>true</tt> or <tt>false</tt></summary>
/// <param name="boolValue">The boolean value to convert</param>
/// <returns>"true" for <tt>true</tt>, "false" for <tt>false</tt></returns>
let internal toLowerBool (boolValue: bool) =
    (string boolValue).ToLowerInvariant()


/// <summary>Valid values for the <tt>hx-swap</tt> attribute / <tt>HX-Reswap</tt> header</summary>
/// <remarks>May be combined with <tt>swap</tt> / <tt>settle</tt> / <tt>scroll</tt> / <tt>show</tt> config)</remarks>
[<RequireQualifiedAccess>]
module HxSwap =
    
    /// <summary>The default, replace the inner HTML of the target element</summary>
    let InnerHtml = "innerHTML"
    
    /// <summary>Replace the entire target element with the response</summary>
    let OuterHtml = "outerHTML"
    
    /// <summary>Insert the response before the target element</summary>
    let BeforeBegin = "beforebegin"
    
    /// <summary>Insert the response before the first child of the target element</summary>
    let AfterBegin = "afterbegin"
    
    /// <summary>Insert the response after the last child of the target element</summary>
    let BeforeEnd = "beforeend"
    
    /// <summary>Insert the response after the target element</summary>
    let AfterEnd = "afterend"
    
    /// <summary>Does not append content from response (out of band items will still be processed).</summary>
    let None = "none"
